@@ -7,24 +7,28 @@
 #
 # @author Davide Brunato <brunato@sissa.it>
 #
-# mypy: ignore-errors
 """
 XPath 1.0 implementation - part 3 (functions)
 """
 import math
 import decimal
+from typing import Any, List, Union
 
-from ..helpers import get_double
-from ..datatypes import Duration, DayTimeDuration, YearMonthDuration, \
-    StringProxy, AnyURI, Float10
-from ..namespaces import XML_ID, XML_LANG, get_prefixed_name
-from ..xpath_nodes import XPathNode, ElementNode, TextNode, CommentNode, \
-    ProcessingInstructionNode, DocumentNode, SchemaElementNode
-from ..xpath_tokens import XPathFunction
-from ..xpath_context import XPathSchemaContext
+from elementpath._typing import Iterator
+from elementpath.aliases import Emptiable
+from elementpath.helpers import get_double
+from elementpath.datatypes import Duration, DayTimeDuration, YearMonthDuration, \
+    StringProxy, AnyURI, Float10, AnyAtomicType, AtomicType, NumericType
+from elementpath.namespaces import XML_ID, XML_LANG
+from elementpath.xpath_nodes import XPathNode, ElementNode, TextNode, CommentNode, \
+    ProcessingInstructionNode, DocumentNode, EtreeElementNode
+from elementpath.xpath_context import ContextType, XPathSchemaContext
+from elementpath.xpath_tokens import XPathFunction
 
 from ._xpath1_operators import XPath1Parser
 
+__all__ = ['XPath1Parser']
+
 method = XPath1Parser.method
 function = XPath1Parser.function
 
@@ -32,7 +36,8 @@ function = XPath1Parser.function
 ###
 # Kind tests (for matching of node types in XPath 1.0 or sequence types in XPath 2.0)
 @method(function('node', nargs=0, label='kind test'))
-def select_node_kind_test(self, context=None):
+def select_node_kind_test(self: XPathFunction, context: ContextType = None) \
+        -> Iterator[XPathNode]:
     if context is None:
         raise self.missing_context()
 
@@ -43,7 +48,7 @@ def select_node_kind_test(self, context=None):
 
 
 @method('node')
-def nud_item_sequence_type(self):
+def nud_item_sequence_type(self: XPathFunction) -> XPathFunction:
     XPathFunction.nud(self)
     if self.parser.next_token.symbol in ('*', '+', '?'):
         self.occurrence = self.parser.next_token.symbol
@@ -52,7 +57,8 @@ def nud_item_sequence_type(self):
 
 
 @method(function('processing-instruction', nargs=(0, 1), bp=79, label='kind test'))
-def select_pi_kind_test(self, context=None):
+def select_pi_kind_test(self: XPathFunction, context: ContextType = None) \
+        -> Iterator[ProcessingInstructionNode]:
     if context is None:
         raise self.missing_context()
 
@@ -62,23 +68,24 @@ def select_pi_kind_test(self, context=None):
                 yield item
             else:
                 name = self[0].value
+                assert isinstance(name, str)
                 if item.name == ' '.join(name.strip().split()):
                     yield item
 
 
 @method('processing-instruction')
-def nud_pi_kind_test(self):
+def nud_pi_kind_test(self: XPathFunction) -> XPathFunction:
     self.parser.advance('(')
     if self.parser.next_token.symbol != ')':
         self.parser.next_token.expected('(name)', '(string)')
         self[0:] = self.parser.expression(5),
     self.parser.advance(')')
-    self.value = None
     return self
 
 
 @method(function('comment', nargs=0, label='kind test'))
-def select_comment_kind_test(self, context=None):
+def select_comment_kind_test(self: XPathFunction, context: ContextType = None) \
+        -> Iterator[CommentNode]:
     if context is None:
         raise self.missing_context()
 
@@ -88,7 +95,8 @@ def select_comment_kind_test(self, context=None):
 
 
 @method(function('text', nargs=0, label='kind test'))
-def select_text_kind_test(self, context=None):
+def select_text_kind_test(self: XPathFunction, context: ContextType = None) \
+        -> Iterator[TextNode]:
     if context is None:
         raise self.missing_context()
 
@@ -100,7 +108,7 @@ def select_text_kind_test(self, context=None):
 ###
 # Node set functions
 @method(function('last', nargs=0, sequence_types=('xs:integer',)))
-def evaluate_last_function(self, context=None):
+def evaluate_last_function(self: XPathFunction, context: ContextType = None) -> int:
     if self.context is not None:
         context = self.context
     elif context is None:
@@ -110,7 +118,7 @@ def evaluate_last_function(self, context=None):
 
 @method(function('position', nargs=0,
                  sequence_types=('xs:integer',)))
-def evaluate_position_function(self, context=None):
+def evaluate_position_function(self: XPathFunction, context: ContextType = None) -> int:
     if self.context is not None:
         context = self.context
     elif context is None:
@@ -119,12 +127,13 @@ def evaluate_position_function(self, context=None):
 
 
 @method(function('count', nargs=1, sequence_types=('item()*', 'xs:integer')))
-def evaluate_count_function(self, context=None):
+def evaluate_count_function(self: XPathFunction, context: ContextType = None) -> int:
     return len([x for x in self[0].select(self.context or context)])
 
 
 @method(function('id', nargs=1, sequence_types=('xs:string*', 'element()*')))
-def select_id_function(self, context=None):
+def select_id_function(self: XPathFunction, context: ContextType = None) \
+        -> Iterator[ElementNode]:
     if self.context is not None:
         context = self.context
     elif context is None:
@@ -137,14 +146,15 @@ def select_id_function(self, context=None):
 
     if isinstance(item, (ElementNode, DocumentNode)):
         for element in item.iter_descendants():
-            if isinstance(element, ElementNode) and element.elem.get(XML_ID) == value:
+            if isinstance(element, EtreeElementNode) and element.obj.get(XML_ID) == value:
                 yield element
 
 
 @method(function('name', nargs=(0, 1), sequence_types=('node()?', 'xs:string')))
 @method(function('local-name', nargs=(0, 1), sequence_types=('node()?', 'xs:string')))
 @method(function('namespace-uri', nargs=(0, 1), sequence_types=('node()?', 'xs:anyURI')))
-def evaluate_name_related_functions(self, context=None):
+def evaluate_name_related_functions(self: XPathFunction, context: ContextType = None) \
+        -> Union[str, AnyURI]:
     if self.context is not None:
         context = self.context
     elif context is None:
@@ -162,8 +172,10 @@ def evaluate_name_related_functions(self, context=None):
 
     symbol = self.symbol
     if symbol == 'name':
-        nsmap = getattr(arg, 'nsmap', self.parser.namespaces)
-        return get_prefixed_name(name, nsmap)
+        node_name = arg.node_name
+        if node_name is None:
+            return ''
+        return node_name.qname
     elif symbol == 'local-name':
         return name if not name or name[0] != '{' else name.split('}')[1]
     elif self.parser.version == '1.0':
@@ -175,7 +187,7 @@ def evaluate_name_related_functions(self, context=None):
 ###
 # String functions
 @method(function('string', nargs=(0, 1), sequence_types=('item()?', 'xs:string')))
-def evaluate_string_function(self, context=None):
+def evaluate_string_function(self: XPathFunction, context: ContextType = None) -> str:
     if self.context is not None:
         context = self.context
 
@@ -188,7 +200,7 @@ def evaluate_string_function(self, context=None):
 
 @method(function('contains', nargs=2,
                  sequence_types=('xs:string?', 'xs:string?', 'xs:boolean')))
-def evaluate_contains_function(self, context=None):
+def evaluate_contains_function(self: XPathFunction, context: ContextType = None) -> bool:
     if self.context is not None:
         context = self.context
 
@@ -199,7 +211,7 @@ def evaluate_contains_function(self, context=None):
 
 @method(function('concat', nargs=(2, None),
                  sequence_types=('xs:anyAtomicType?', 'xs:anyAtomicType?', 'xs:string')))
-def evaluate_concat_function(self, context=None):
+def evaluate_concat_function(self: XPathFunction, context: ContextType = None) -> str:
     if self.context is not None:
         context = self.context
 
@@ -210,7 +222,7 @@ def evaluate_concat_function(self, context=None):
 
 @method(function('string-length', nargs=(0, 1),
                  sequence_types=('xs:string?', 'xs:integer')))
-def evaluate_string_length_function(self, context=None):
+def evaluate_string_length_function(self: XPathFunction, context: ContextType = None) -> int:
     if self.context is not None:
         context = self.context
 
@@ -224,7 +236,7 @@ def evaluate_string_length_function(self, context=None):
 
 @method(function('normalize-space', nargs=(0, 1),
                  sequence_types=('xs:string?', 'xs:string')))
-def evaluate_normalize_space_function(self, context=None):
+def evaluate_normalize_space_function(self: XPathFunction, context: ContextType = None) -> str:
     if self.context is not None:
         context = self.context
 
@@ -237,29 +249,29 @@ def evaluate_normalize_space_function(self, context=None):
 
 @method(function('starts-with', nargs=2,
                  sequence_types=('xs:string?', 'xs:string?', 'xs:boolean')))
-def evaluate_starts_with_function(self, context=None):
+def evaluate_starts_with_function(self: XPathFunction, context: ContextType = None) -> bool:
     if self.context is not None:
         context = self.context
 
-    arg1 = self.get_argument(context, default='', cls=str)
-    arg2 = self.get_argument(context, index=1, default='', cls=str)
+    arg1: str = self.get_argument(context, default='', cls=str)
+    arg2: str = self.get_argument(context, index=1, default='', cls=str)
     return arg1.startswith(arg2)
 
 
 @method(function('translate', nargs=3,
                  sequence_types=('xs:string?', 'xs:string', 'xs:string', 'xs:string')))
-def evaluate_translate_function(self, context=None):
+def evaluate_translate_function(self: XPathFunction, context: ContextType = None) -> str:
     if self.context is not None:
         context = self.context
 
-    arg = self.get_argument(context, default='', cls=str)
+    arg: str = self.get_argument(context, default='', cls=str)
 
-    map_string = self.get_argument(context, index=1, cls=str)
+    map_string: str = self.get_argument(context, index=1, cls=str)
     if map_string is None:
         message = "the 2nd argument of fn:translate() cannot be the empty sequence"
         raise self.error('XPTY0004', message)
 
-    trans_string = self.get_argument(context, index=2, cls=str)
+    trans_string: str = self.get_argument(context, index=2, cls=str)
     if trans_string is None:
         message = "the 3rd argument of fn:translate() cannot be the empty sequence"
         raise self.error('XPTY0004', message)
@@ -275,11 +287,11 @@ def evaluate_translate_function(self, context=None):
 
 @method(function('substring', nargs=(2, 3),
                  sequence_types=('xs:string?', 'xs:double', 'xs:double', 'xs:string')))
-def evaluate_substring_function(self, context=None):
+def evaluate_substring_function(self: XPathFunction, context: ContextType = None) -> str:
     if self.context is not None:
         context = self.context
 
-    item = self.get_argument(context, default='', cls=str)
+    item: str = self.get_argument(context, default='', cls=str)
     try:
         start = self.get_argument(context, index=1, required=True)
         if math.isnan(start) or math.isinf(start):
@@ -316,12 +328,13 @@ def evaluate_substring_function(self, context=None):
                  sequence_types=('xs:string?', 'xs:string?', 'xs:string')))
 @method(function('substring-after', nargs=2,
                  sequence_types=('xs:string?', 'xs:string?', 'xs:string')))
-def evaluate_substring_before_or_after_functions(self, context=None):
+def evaluate_substring_before_or_after_functions(
+        self: XPathFunction, context: ContextType = None) -> str:
     if self.context is not None:
         context = self.context
 
-    arg1 = self.get_argument(context, default='', cls=str)
-    arg2 = self.get_argument(context, index=1, default='', cls=str)
+    arg1: str = self.get_argument(context, default='', cls=str)
+    arg2: str = self.get_argument(context, index=1, default='', cls=str)
 
     index = arg1.find(arg2)
     if index < 0:
@@ -336,43 +349,44 @@ def evaluate_substring_before_or_after_functions(self, context=None):
 # Boolean functions
 @method(function('boolean', nargs=1,
                  sequence_types=('item()*', 'xs:boolean')))
-def evaluate_boolean_function(self, context=None):
-    return self.boolean_value([x for x in self[0].select(self.context or context)])
+def evaluate_boolean_function(self: XPathFunction, context: ContextType = None) -> bool:
+    return self.boolean_value(self[0].select(self.context or context))
 
 
 @method(function('not', nargs=1, sequence_types=('item()*', 'xs:boolean')))
-def evaluate_not_function(self, context=None):
-    return not self.boolean_value([x for x in self[0].select(self.context or context)])
+def evaluate_not_function(self: XPathFunction, context: ContextType = None) -> bool:
+    return not self.boolean_value(self[0].select(self.context or context))
 
 
 @method(function('true', nargs=0, sequence_types=('xs:boolean',)))
-def evaluate_true_function(self, context=None):
+def evaluate_true_function(self: XPathFunction, context: ContextType = None) -> bool:
     return True
 
 
 @method(function('false', nargs=0, sequence_types=('xs:boolean',)))
-def evaluate_false_function(self, context=None):
+def evaluate_false_function(self: XPathFunction, context: ContextType = None) -> bool:
     return False
 
 
 @method(function('lang', nargs=1,
                  sequence_types=('xs:string?', 'xs:boolean')))
-def evaluate_lang_function(self, context=None):
+def evaluate_lang_function(self: XPathFunction, context: ContextType = None) -> bool:
     if self.context is not None:
         context = self.context
     elif context is None:
         raise self.missing_context()
 
-    if not isinstance(context.item, ElementNode) \
-            or isinstance(context.item, SchemaElementNode):
+    if not isinstance(context.item, EtreeElementNode):
         return False
     else:
         try:
-            attr = context.item.elem.attrib[XML_LANG]
+            attr = context.item.obj.attrib[XML_LANG]
         except KeyError:
             for e in context.iter_ancestors():
-                if isinstance(e, ElementNode) and XML_LANG in e.elem.attrib:
-                    lang = e.elem.attrib[XML_LANG]
+                if isinstance(e, EtreeElementNode) and XML_LANG in e.obj.attrib:
+                    lang = e.obj.attrib[XML_LANG]
+                    if not isinstance(lang, str):
+                        return False
                     break
             else:
                 return False
@@ -383,28 +397,34 @@ def evaluate_lang_function(self, context=None):
 
         if '-' in lang:
             lang, _ = lang.split('-')
-        return lang.lower() == self[0].evaluate().lower()
+
+        value = self[0].evaluate()
+        if not isinstance(value, str):
+            return False
+        return lang.lower() == value.lower()
 
 
 ###
 # Number functions
 @method(function('number', nargs=(0, 1), sequence_types=('xs:anyAtomicType?', 'xs:double')))
-def evaluate_number_function(self, context=None):
+def evaluate_number_function(self: XPathFunction, context: ContextType = None) -> float:
     arg = self.get_argument(self.context or context, default_to_context=True)
     return self.number_value(arg)
 
 
 @method(function('sum', nargs=(1, 2),
                  sequence_types=('xs:anyAtomicType*', 'xs:anyAtomicType?', 'xs:anyAtomicType?')))
-def evaluate_sum_function(self, context=None):
+def evaluate_sum_function(self: XPathFunction, context: ContextType = None) \
+        -> Emptiable[AtomicType]:
     if self.context is not None:
         context = self.context
 
     xsd_version = self.parser.xsd_version
+    values: List[Any]
     try:
         values = [get_double(self.string_value(x), xsd_version)
                   if isinstance(x, XPathNode) else x
-                  for x in self[0].iter_flatten(context)]
+                  for x in self[0].select_flatten(context)]
     except (TypeError, ValueError):
         if self.parser.version == '1.0':
             return math.nan
@@ -417,10 +437,10 @@ def evaluate_sum_function(self, context=None):
         return [] if zero is None else zero
 
     if all(isinstance(x, (decimal.Decimal, int)) for x in values):
-        return sum(values) if len(values) > 1 else values[0]
+        result = sum(values) if len(values) > 1 else values[0]
     elif all(isinstance(x, DayTimeDuration) for x in values) or \
             all(isinstance(x, YearMonthDuration) for x in values):
-        return sum(values[1:], start=values[0])
+        result = sum(values[1:], start=values[0])
     elif any(isinstance(x, Duration) for x in values):
         raise self.error('FORG0006', 'invalid sum of duration values')
     elif any(isinstance(x, (StringProxy, AnyURI)) for x in values):
@@ -428,21 +448,25 @@ def evaluate_sum_function(self, context=None):
     elif any(isinstance(x, float) and math.isnan(x) for x in values):
         return math.nan
     elif all(isinstance(x, Float10) for x in values):
-        return sum(values)
+        result = sum(values)
+    else:
+        try:
+            result = sum(self.number_value(x) for x in values)
+        except TypeError:
+            if self.parser.version == '1.0':
+                return math.nan
+            elif isinstance(context, XPathSchemaContext):
+                return []
+            raise self.error('FORG0006') from None
 
-    try:
-        return sum(self.number_value(x) for x in values)
-    except TypeError:
-        if self.parser.version == '1.0':
-            return math.nan
-        elif isinstance(context, XPathSchemaContext):
-            return []
-        raise self.error('FORG0006') from None
+    assert isinstance(result, AnyAtomicType)
+    return result
 
 
 @method(function('ceiling', nargs=1, sequence_types=('xs:numeric?', 'xs:numeric?')))
 @method(function('floor', nargs=1, sequence_types=('xs:numeric?', 'xs:numeric?')))
-def evaluate_ceiling_and_floor_functions(self, context=None):
+def evaluate_ceiling_and_floor_functions(self: XPathFunction, context: ContextType = None) \
+        -> Emptiable[NumericType]:
     if self.context is not None:
         context = self.context
 
@@ -454,8 +478,10 @@ def evaluate_ceiling_and_floor_functions(self, context=None):
 
     try:
         if math.isnan(arg) or math.isinf(arg):
+            assert isinstance(arg, (int, float, decimal.Decimal))
             return arg
 
+        assert isinstance(arg, (int, float, decimal.Decimal))
         if self.symbol == 'floor':
             return type(arg)(math.floor(arg))
         else:
@@ -469,7 +495,8 @@ def evaluate_ceiling_and_floor_functions(self, context=None):
 
 
 @method(function('round', nargs=1, sequence_types=('xs:numeric?', 'xs:numeric?')))
-def evaluate_round_function(self, context=None):
+def evaluate_round_function(self: XPathFunction, context: ContextType = None) \
+        -> Emptiable[NumericType]:
     if self.context is not None:
         context = self.context
 
@@ -484,6 +511,7 @@ def evaluate_round_function(self, context=None):
 
     try:
         number = decimal.Decimal(arg)
+        assert isinstance(arg, (int, float, decimal.Decimal))
         if number > 0:
             return type(arg)(number.quantize(decimal.Decimal('1'), rounding='ROUND_HALF_UP'))
         else:
@@ -494,6 +522,7 @@ def evaluate_round_function(self, context=None):
         raise self.error('FORG0006', err) from None
     except decimal.InvalidOperation:
         if not isinstance(arg, str):
+            assert isinstance(arg, (int, float, decimal.Decimal))
             return round(arg)
         elif isinstance(context, XPathSchemaContext):
             return []
