@@ -7,7 +7,6 @@
 #
 # @author Davide Brunato <brunato@sissa.it>
 #
-# mypy: ignore-errors
 """
 XPath 1.0 implementation - part 2 (operators and expressions)
 """
@@ -15,20 +14,23 @@ import math
 import decimal
 import operator
 from copy import copy
+from typing import Any, cast, List, NoReturn, Optional, Set, Type, Union
 
-from ..datatypes import AnyURI
-from ..exceptions import ElementPathKeyError, ElementPathTypeError
-from ..helpers import collapse_white_spaces, node_position
-from ..datatypes import AbstractDateTime, Duration, DayTimeDuration, \
-    YearMonthDuration, NumericProxy, ArithmeticProxy
-from ..xpath_context import XPathSchemaContext
-from ..namespaces import XMLNS_NAMESPACE, XSD_NAMESPACE
-from ..schema_proxy import AbstractSchemaProxy
-from ..xpath_nodes import XPathNode, ElementNode, AttributeNode, DocumentNode
-from ..xpath_tokens import XPathToken
+from elementpath._typing import Iterator, Sequence
+from elementpath.exceptions import ElementPathKeyError, ElementPathTypeError
+from elementpath.helpers import collapse_white_spaces, node_position
+from elementpath.datatypes import AbstractDateTime, AnyURI, Duration, DayTimeDuration, \
+    YearMonthDuration, NumericProxy, ArithmeticProxy, NumericType, ArithmeticType
+from elementpath.xpath_context import ContextType, ItemType, XPathSchemaContext
+from elementpath.namespaces import XMLNS_NAMESPACE, XSD_NAMESPACE
+from elementpath.xpath_nodes import ParentNodeType, XPathNode, \
+    ElementNode, AttributeNode, DocumentNode
+from elementpath.xpath_tokens import XPathParserType, XPathToken, XPathTokenType
 
 from .xpath1_parser import XPath1Parser
 
+__all__ = ['XPath1Parser']
+
 OPERATORS_MAP = {
     '=': operator.eq,
     '!=': operator.ne,
@@ -40,16 +42,12 @@ OPERATORS_MAP = {
 
 register = XPath1Parser.register
 nullary = XPath1Parser.nullary
-prefix = XPath1Parser.prefix
 infix = XPath1Parser.infix
-postfix = XPath1Parser.postfix
 method = XPath1Parser.method
-function = XPath1Parser.function
-axis = XPath1Parser.axis
 
 
 @method(register('(name)', bp=10, label='literal'))
-def nud_name_literal(self):
+def nud_name_literal(self: XPathToken) -> XPathToken:
     if self.parser.next_token.symbol == '::':
         msg = "axis '%s::' not found" % self.value
         if self.parser.compatibility_mode:
@@ -70,60 +68,19 @@ def nud_name_literal(self):
 
 
 @method('(name)')
-def evaluate_name_literal(self, context=None):
+def evaluate_name_literal(self: XPathToken, context: ContextType = None) \
+        -> List[ItemType]:
     return [x for x in self.select(context)]
 
 
 @method('(name)')
-def select_name_literal(self, context=None):
+def select_name_literal(self: XPathToken, context: ContextType = None) \
+        -> Iterator[ItemType]:
     if context is None:
         raise self.missing_context()
-    elif isinstance(context, XPathSchemaContext):
-        yield from self.select_xsd_nodes(context, self.value)
-        return
-    else:
-        name = self.value
-        default_namespace = self.parser.default_namespace
 
-    # With an ElementTree context checks if the token is bound to an XSD type. If not
-    # try a match using the element path. If this match fails the xsd_type attribute
-    # is set with the schema object to prevent other checks until the schema change.
-    if self.xsd_types is self.parser.schema:
-
-        # Untyped selection
-        for item in context.iter_children_or_self():
-            if item.match_name(name, default_namespace):
-                yield item
-
-    elif self.xsd_types is None or isinstance(self.xsd_types, AbstractSchemaProxy):
-
-        # Try to match the type using the item's path
-        for item in context.iter_children_or_self():
-            if item.match_name(name, default_namespace):
-
-                if item.xsd_type is not None:
-                    yield item
-                else:
-                    xsd_node = self.parser.schema.find(item.path, self.parser.namespaces)
-                    if xsd_node is None:
-                        self.xsd_types = self.parser.schema
-                    elif isinstance(item, AttributeNode):
-                        self.xsd_types = {item.name: xsd_node.type}
-                    else:
-                        self.xsd_types = {item.elem.tag: xsd_node.type}
-
-                    context.item = self.get_typed_node(item)
-                    yield context.item
-
-    else:
-        # XSD typed selection
-        for item in context.iter_children_or_self():
-            if item.match_name(name, default_namespace):
-                if item.xsd_type is not None:
-                    yield item
-                else:
-                    context.item = self.get_typed_node(item)
-                    yield context.item
+    if isinstance(self.value, str):
+        yield from context.iter_matching_nodes(self.value, self.parser.default_namespace)
 
 
 ###
@@ -134,8 +91,9 @@ class _PrefixedReferenceToken(XPathToken):
     symbol = lookup_name = ':'
     lbp = 95
     rbp = 95
+    value: str
 
-    def __init__(self, parser, value=None):
+    def __init__(self, parser: XPathParserType, value: Optional[Any] = None) -> None:
         super().__init__(parser, value)
 
         # Change bind powers if it cannot be a namespace related token
@@ -144,7 +102,7 @@ class _PrefixedReferenceToken(XPathToken):
         elif self.parser.token.symbol not in ('*', '(name)', 'array'):
             self.lbp = self.rbp = 0
 
-    def __str__(self):
+    def __str__(self) -> str:
         if len(self) < 2:
             return 'unparsed prefixed reference'
         elif self[1].label.endswith('function'):
@@ -155,13 +113,22 @@ class _PrefixedReferenceToken(XPathToken):
             return f"{self.value!r} prefixed name"
 
     @property
-    def source(self) -> str:
+    def source(self: XPathToken) -> str:
         if self.occurrence:
             return ':'.join(tk.source for tk in self) + self.occurrence
         else:
             return ':'.join(tk.source for tk in self)
 
-    def led(self, left):
+    @property
+    def name(self) -> str:
+        prefix = self[0].value
+        assert isinstance(prefix, str)
+        if prefix == '*':
+            return '*:%s' % self[1].value
+        else:
+            return f'{{{self.get_namespace(prefix)}}}{self[1].value}'
+
+    def led(self: XPathToken, left: XPathToken) -> XPathToken:
         version = self.parser.version
         if self.is_spaced():
             if version <= '3.0':
@@ -180,7 +147,7 @@ class _PrefixedReferenceToken(XPathToken):
 
         if left.symbol == '(name)':
             try:
-                namespace = self.get_namespace(left.value)
+                namespace = self.get_namespace(cast(str, left.value))
             except ElementPathKeyError:
                 self.parser.advance()  # Assure there isn't a following incomplete comment
                 self[:] = left, self.parser.token
@@ -192,18 +159,20 @@ class _PrefixedReferenceToken(XPathToken):
             raise self.wrong_syntax()
 
         self[:] = left, self.parser.expression(95)
+
         if self[1].label.endswith('function'):
             self.value = f'{self[0].value}:{self[1].symbol}'
         else:
             self.value = f'{self[0].value}:{self[1].value}'
         return self
 
-    def evaluate(self, context=None):
+    def evaluate(self: XPathToken, context: ContextType = None) \
+            -> Union[ItemType, List[ItemType]]:
         if self[1].label.endswith('function'):
             return self[1].evaluate(context)
         return [x for x in self.select(context)]
 
-    def select(self, context=None):
+    def select(self, context: ContextType = None) -> Iterator[ItemType]:
         if self[1].label.endswith('function'):
             value = self[1].evaluate(context)
             if isinstance(value, list):
@@ -212,47 +181,10 @@ class _PrefixedReferenceToken(XPathToken):
                 yield value
             return
 
-        if self[0].value == '*':
-            name = '*:%s' % self[1].value
-        else:
-            name = '{%s}%s' % (self.get_namespace(self[0].value), self[1].value)
-
         if context is None:
             raise self.missing_context()
-        elif isinstance(context, XPathSchemaContext):
-            yield from self.select_xsd_nodes(context, name)
 
-        elif self.xsd_types is self.parser.schema:
-            for item in context.iter_children_or_self():
-                if item.match_name(name):
-                    yield item
-
-        elif self.xsd_types is None or isinstance(self.xsd_types, AbstractSchemaProxy):
-            for item in context.iter_children_or_self():
-                if item.match_name(name):
-                    assert isinstance(item, (ElementNode, AttributeNode))
-                    if item.xsd_type is not None:
-                        yield item
-                    else:
-                        xsd_node = self.parser.schema.find(item.path, self.parser.namespaces)
-                        if xsd_node is not None:
-                            self.add_xsd_type(xsd_node)
-                        else:
-                            self.xsd_types = self.parser.schema
-
-                        context.item = self.get_typed_node(item)
-                        yield context.item
-
-        else:
-            # XSD typed selection
-            for item in context.iter_children_or_self():
-                if item.match_name(name):
-                    assert isinstance(item, (ElementNode, AttributeNode))
-                    if item.xsd_type is not None:
-                        yield item
-                    else:
-                        context.item = self.get_typed_node(item)
-                        yield context.item
+        yield from context.iter_matching_nodes(self.name)
 
 
 XPath1Parser.symbol_table[':'] = _PrefixedReferenceToken
@@ -261,7 +193,7 @@ XPath1Parser.symbol_table[':'] = _PrefixedReferenceToken
 ###
 # Namespace URI as in ElementPath
 @method('{', bp=95)
-def nud_namespace_uri(self):
+def nud_namespace_uri(self: XPathToken) -> XPathToken:
     if self.parser.strict and self.symbol == '{':
         raise self.wrong_syntax("not allowed symbol if parser has strict=True")
 
@@ -269,7 +201,9 @@ def nud_namespace_uri(self):
     if self.parser.next_token.symbol == '}':
         namespace = ''
     else:
-        namespace = self.parser.next_token.value + self.parser.advance_until('}')
+        value = self.parser.next_token.value
+        assert isinstance(value, str)
+        namespace = value + self.parser.advance_until('}')
         namespace = collapse_white_spaces(namespace)
 
     try:
@@ -287,68 +221,56 @@ def nud_namespace_uri(self):
         self.parser.expected_next('(name)', '*')
     self.parser.next_token.bind_namespace(namespace)
 
-    self[:] = self.parser.symbol_table['(string)'](self.parser, namespace), \
-        self.parser.expression(90)
+    cls: Type[XPathToken] = self.parser.symbol_table['(string)']
+    self[:] = cls(self.parser, namespace), self.parser.expression(90)
 
-    if self[1].value is None or not self[0].value:
+    if not self[0].value:
         self.value = self[1].value
     else:
-        self.value = '{%s}%s' % (self[0].value, self[1].value)
+        self.value = f'{{{self[0].value}}}{self[1].value}'
     return self
 
 
 @method('{')
-def evaluate_namespace_uri(self, context=None):
+def evaluate_namespace_uri(self: XPathToken, context: ContextType = None) \
+        -> Union[ItemType, List[ItemType]]:
     if self[1].label.endswith('function'):
         return self[1].evaluate(context)
     return [x for x in self.select(context)]
 
 
 @method('{')
-def select_namespace_uri(self, context=None):
+def select_namespace_uri(self: XPathToken, context: ContextType = None) \
+        -> Iterator[Union[ItemType, List[ItemType]]]:
     if self[1].label.endswith('function'):
         yield self[1].evaluate(context)
         return
     elif context is None:
         raise self.missing_context()
 
-    if isinstance(context, XPathSchemaContext):
-        yield from self.select_xsd_nodes(context, self.value)
-
-    elif self.xsd_types is None:
-        for item in context.iter_children_or_self():
-            if item.match_name(self.value):
-                yield item
-    else:
-        # XSD typed selection
-        for item in context.iter_children_or_self():
-            if item.match_name(self.value):
-                assert isinstance(item, (ElementNode, AttributeNode))
-                if item.xsd_type is not None:
-                    yield item
-                else:
-                    context.item = self.get_typed_node(item)
-                    yield context.item
+    if isinstance(self.value, str):
+        yield from context.iter_matching_nodes(self.value)
 
 
 ###
 # Variables
 @method('$', bp=90)
-def nud_variable_reference(self):
+def nud_variable_reference(self: XPathToken) -> XPathToken:
     self.parser.expected_next('(name)')
     self[:] = self.parser.expression(rbp=90),
-    if ':' in self[0].value:
+    if not isinstance(self[0].value, str) or ':' in self[0].value:
         raise self[0].wrong_syntax("variable reference requires a simple reference name")
     return self
 
 
 @method('$')
-def evaluate_variable_reference(self, context=None):
+def evaluate_variable_reference(self: XPathToken, context: ContextType = None) \
+        -> Union[ItemType, List[ItemType]]:
     if context is None:
         raise self.missing_context()
 
     try:
-        value = context.variables[self[0].value]
+        value = context.variables[cast(str, self[0].value)]
     except KeyError as err:
         raise self.error('XPST0008', 'unknown variable %r' % str(err)) from None
     else:
@@ -358,25 +280,25 @@ def evaluate_variable_reference(self, context=None):
 ###
 # Nullary operators (use only the context)
 @method(nullary('*'))
-def select_wildcard(self, context=None):
+def select_wildcard(self: XPathToken, context: ContextType = None) -> Iterator[ItemType]:
     if self:
         # Product operator
         item = self.evaluate(context)
-        if item or not isinstance(item, list):
+        if not isinstance(item, list):
             if context is not None:
                 context.item = item
             yield item
+        elif context is not None:
+            for context.item in item:
+                yield context.item
+        else:
+            yield from item
+        return
     elif context is None:
         raise self.missing_context()
 
     # Wildcard literal
-    elif isinstance(context, XPathSchemaContext):
-        for item in context.iter_children_or_self():
-            if item is not None:
-                self.add_xsd_type(item)
-                yield item
-
-    elif self.xsd_types is None:
+    if self.parser.schema is None:
         for item in context.iter_children_or_self():
             if item is None:
                 pass  # '*' wildcard doesn't match document nodes
@@ -385,57 +307,24 @@ def select_wildcard(self, context=None):
                     yield item
             elif isinstance(item, ElementNode):
                 yield item
-
     else:
         # XSD typed selection
         for item in context.iter_children_or_self():
-            if context.item is not None and context.is_principal_node_kind():
-                if isinstance(item, (ElementNode, AttributeNode)) and \
-                        item.xsd_type is not None:
+            if context.is_principal_node_kind():
+                if isinstance(item, (ElementNode, AttributeNode)):
                     yield item
-                else:
-                    context.item = self.get_typed_node(item)
-                    yield context.item
 
 
 @method(nullary('.'))
-def select_self_shortcut(self, context=None):
+def select_self_shortcut(self: XPathToken, context: ContextType = None) -> Iterator[ItemType]:
     if context is None:
         raise self.missing_context()
-
-    elif isinstance(context, XPathSchemaContext):
-        for item in context.iter_self():
-            if isinstance(item, (AttributeNode, ElementNode)):
-                if item.is_schema_node():
-                    self.add_xsd_type(item)
-                elif item is context.root:
-                    # item is the schema
-                    for xsd_element in item:
-                        self.add_xsd_type(xsd_element)
-            yield item
-
-    elif self.xsd_types is None:
-        for item in context.iter_self():
-            if item is not None:
-                yield item
-            elif isinstance(context.root, DocumentNode):
-                yield context.root
-
-    else:
-        for item in context.iter_self():
-            if item is not None:
-                if isinstance(item, (ElementNode, AttributeNode)) and \
-                        item.xsd_type is not None:
-                    yield item
-                else:
-                    context.item = self.get_typed_node(item)
-                    yield context.item
-            elif isinstance(context.root, DocumentNode):
-                yield context.root
+    yield from context.iter_self()
 
 
 @method(nullary('..'))
-def select_parent_shortcut(self, context=None):
+def select_parent_shortcut(self: XPathToken, context: ContextType = None) \
+        -> Iterator[ParentNodeType]:
     if context is None:
         raise self.missing_context()
     yield from context.iter_parent()
@@ -444,15 +333,25 @@ def select_parent_shortcut(self, context=None):
 ###
 # Logical Operators
 @method(infix('or', bp=20))
-def evaluate_or_operator(self, context=None):
-    return self.boolean_value(self[0].evaluate(copy(context))) or \
-        self.boolean_value(self[1].evaluate(copy(context)))
+def evaluate_or_operator(self: XPathToken, context: ContextType = None) -> bool:
+    if isinstance(context, XPathSchemaContext):
+        op1 = self.boolean_value(self[0].select(copy(context)))
+        op2 = self.boolean_value(self[1].select(copy(context)))
+        return op1 or op2
+
+    return self.boolean_value(self[0].select(copy(context))) or \
+        self.boolean_value(self[1].select(copy(context)))
 
 
 @method(infix('and', bp=25))
-def evaluate_and_operator(self, context=None):
-    return self.boolean_value(self[0].evaluate(copy(context))) and \
-        self.boolean_value(self[1].evaluate(copy(context)))
+def evaluate_and_operator(self: XPathToken, context: ContextType = None) -> bool:
+    if isinstance(context, XPathSchemaContext):
+        op1 = self.boolean_value(self[0].select(copy(context)))
+        op2 = self.boolean_value(self[1].select(copy(context)))
+        return op1 and op2
+
+    return self.boolean_value(self[0].select(copy(context))) and \
+        self.boolean_value(self[1].select(copy(context)))
 
 
 ###
@@ -463,7 +362,7 @@ def evaluate_and_operator(self, context=None):
 @method('>', bp=30)
 @method('<=', bp=30)
 @method('>=', bp=30)
-def led_comparison_operators(self, left):
+def led_comparison_operators(self: XPathToken, left: XPathToken) -> XPathToken:
     if left.symbol in OPERATORS_MAP:
         raise self.wrong_syntax()
     self[:] = left, self.parser.expression(rbp=30)
@@ -476,36 +375,44 @@ def led_comparison_operators(self, left):
 @method('>')
 @method('<=')
 @method('>=')
-def evaluate_comparison_operators(self, context=None):
+def evaluate_comparison_operators(self: XPathToken, context: ContextType = None) -> bool:
     op = OPERATORS_MAP[self.symbol]
     try:
         return any(op(x1, x2) for x1, x2 in self.iter_comparison_data(context))
-    except ElementPathTypeError:
-        raise
-    except TypeError as err:
-        raise self.error('XPTY0004', err) from None
-    except ValueError as err:
-        raise self.error('FORG0001', err) from None
+    except (TypeError, ValueError) as err:
+        if isinstance(context, XPathSchemaContext):
+            return False
+        elif isinstance(err, ElementPathTypeError):
+            raise
+        elif isinstance(err, TypeError):
+            raise self.error('XPTY0004', err) from None
+        else:
+            raise self.error('FORG0001', err) from None
 
 
 ###
 # Numerical operators
 @method(infix('+', bp=40))
-def evaluate_plus_operator(self, context=None):
+def evaluate_plus_operator(self: XPathToken, context: ContextType = None) \
+        -> Union[List[NoReturn], ArithmeticType]:
     if len(self) == 1:
-        arg = self.get_argument(context, cls=NumericProxy)
+        arg: NumericType = self.get_argument(context, cls=NumericProxy)
         return [] if arg is None else +arg
     else:
+        op1: Optional[ArithmeticType]
+        op2: ArithmeticType
         op1, op2 = self.get_operands(context, cls=ArithmeticProxy)
         if op1 is None:
             return []
 
         try:
-            return op1 + op2
-        except TypeError as err:
-            raise self.error('XPTY0004', err) from None
-        except OverflowError as err:
-            if isinstance(op1, AbstractDateTime):
+            return op1 + op2  # type:ignore[operator]
+        except (TypeError, OverflowError) as err:
+            if isinstance(context, XPathSchemaContext):
+                return []
+            elif isinstance(err, TypeError):
+                raise self.error('XPTY0004', err) from None
+            elif isinstance(op1, AbstractDateTime):
                 raise self.error('FODT0001', err) from None
             elif isinstance(op1, Duration):
                 raise self.error('FODT0002', err) from None
@@ -514,21 +421,26 @@ def evaluate_plus_operator(self, context=None):
 
 
 @method(infix('-', bp=40))
-def evaluate_minus_operator(self, context=None):
+def evaluate_minus_operator(self: XPathToken, context: ContextType = None) \
+        -> Union[List[NoReturn], ArithmeticType]:
     if len(self) == 1:
-        arg = self.get_argument(context, cls=NumericProxy)
+        arg: NumericType = self.get_argument(context, cls=NumericProxy)
         return [] if arg is None else -arg
     else:
+        op1: Optional[ArithmeticType]
+        op2: ArithmeticType
         op1, op2 = self.get_operands(context, cls=ArithmeticProxy)
         if op1 is None:
             return []
 
         try:
-            return op1 - op2
-        except TypeError as err:
-            raise self.error('XPTY0004', err) from None
-        except OverflowError as err:
-            if isinstance(op1, AbstractDateTime):
+            return op1 - op2  # type:ignore[operator]
+        except (TypeError, OverflowError) as err:
+            if isinstance(context, XPathSchemaContext):
+                return []
+            elif isinstance(err, TypeError):
+                raise self.error('XPTY0004', err) from None
+            elif isinstance(op1, AbstractDateTime):
                 raise self.error('FODT0001', err) from None
             elif isinstance(op1, Duration):
                 raise self.error('FODT0002', err) from None
@@ -538,13 +450,16 @@ def evaluate_minus_operator(self, context=None):
 
 @method('+')
 @method('-')
-def nud_plus_minus_operators(self):
+def nud_plus_minus_operators(self: XPathToken) -> XPathToken:
     self[:] = self.parser.expression(rbp=70),
     return self
 
 
 @method(infix('*', bp=45))
-def evaluate_multiply_operator(self, context=None):
+def evaluate_multiply_operator(self: XPathToken, context: ContextType = None) \
+        -> Union[ArithmeticType, List[ItemType]]:
+    op1: Optional[ArithmeticType]
+    op2: ArithmeticType
     if self:
         op1, op2 = self.get_operands(context, cls=ArithmeticProxy)
         if op1 is None:
@@ -552,8 +467,11 @@ def evaluate_multiply_operator(self, context=None):
         try:
             if isinstance(op2, (YearMonthDuration, DayTimeDuration)):
                 return op2 * op1
-            return op1 * op2
+            return op1 * op2  # type:ignore[operator]
         except TypeError as err:
+            if isinstance(context, XPathSchemaContext):
+                return []
+
             if isinstance(op1, (float, decimal.Decimal)):
                 if math.isnan(op1):
                     raise self.error('FOCA0005') from None
@@ -568,9 +486,13 @@ def evaluate_multiply_operator(self, context=None):
 
             raise self.error('XPTY0004', err) from None
         except ValueError as err:
+            if isinstance(context, XPathSchemaContext):
+                return []
             raise self.error('FOCA0005', err) from None
         except OverflowError as err:
-            if isinstance(op1, AbstractDateTime):
+            if isinstance(context, XPathSchemaContext):
+                return []
+            elif isinstance(op1, AbstractDateTime):
                 raise self.error('FODT0001', err) from None
             elif isinstance(op1, Duration):
                 raise self.error('FODT0002', err) from None
@@ -582,7 +504,10 @@ def evaluate_multiply_operator(self, context=None):
 
 
 @method(infix('div', bp=45))
-def evaluate_div_operator(self, context=None):
+def evaluate_div_operator(self: XPathToken, context: ContextType = None) \
+        -> Union[int, float, decimal.Decimal, List[Any]]:
+    dividend: Optional[ArithmeticType]
+    divisor: ArithmeticType
     dividend, divisor = self.get_operands(context, cls=ArithmeticProxy)
     if dividend is None:
         return []
@@ -590,7 +515,7 @@ def evaluate_div_operator(self, context=None):
         try:
             if isinstance(dividend, int) and isinstance(divisor, int):
                 return decimal.Decimal(dividend) / decimal.Decimal(divisor)
-            return dividend / divisor
+            return dividend / divisor  # type:ignore[operator]
         except TypeError as err:
             raise self.error('XPTY0004', err) from None
         except ValueError as err:
@@ -617,10 +542,15 @@ def evaluate_div_operator(self, context=None):
 
 
 @method(infix('mod', bp=45))
-def evaluate_mod_operator(self, context=None):
+def evaluate_mod_operator(self: XPathToken, context: ContextType = None) \
+        -> Union[List[NoReturn], ArithmeticType]:
+    op1: Optional[NumericType]
+    op2: Optional[NumericType]
     op1, op2 = self.get_operands(context, cls=NumericProxy)
     if op1 is None:
         return []
+    elif op2 is None:
+        raise self.error('XPTY0004', '2nd operand is an empty sequence')
     elif op2 == 0 and isinstance(op2, float):
         return math.nan
     elif math.isinf(op2) and not math.isinf(op1) and op1 != 0:
@@ -629,7 +559,7 @@ def evaluate_mod_operator(self, context=None):
     try:
         if isinstance(op1, int) and isinstance(op2, int):
             return op1 % op2 if op1 * op2 >= 0 else -(abs(op1) % op2)
-        return op1 % op2
+        return op1 % op2  # type: ignore[operator]
     except TypeError as err:
         raise self.error('FORG0006', err) from None
     except (ZeroDivisionError, decimal.InvalidOperation):
@@ -641,39 +571,39 @@ def evaluate_mod_operator(self, context=None):
 @method('and')
 @method('div')
 @method('mod')
-def nud_disambiguation_of_infix_operators(self):
+def nud_disambiguation_of_infix_operators(self: XPathToken) -> XPathTokenType:
     return self.as_name()
 
 
 ###
 # Union expressions
 @method('|', bp=50)
-def led_union_operator(self, left):
-    self.cut_and_sort = True
+def led_union_operator(self: XPathToken, left: XPathToken) -> XPathToken:
     if left.symbol in ('|', 'union'):
-        left.cut_and_sort = False
+        left.concatenated = True
     self[:] = left, self.parser.expression(rbp=50)
     return self
 
 
 @method('|')
-def select_union_operator(self, context=None):
+def select_union_operator(self: XPathToken, context: ContextType = None) \
+        -> Iterator[XPathNode]:
     if context is None:
         raise self.missing_context()
 
     results = {item for k in range(2) for item in self[k].select(copy(context))}
     if any(not isinstance(x, XPathNode) for x in results):
         raise self.error('XPTY0004', 'only XPath nodes are allowed')
-    elif not self.cut_and_sort:
-        yield from results
+    elif self.concatenated:
+        yield from cast(Set[XPathNode], results)
     else:
-        yield from sorted(results, key=node_position)
+        yield from cast(List[XPathNode], sorted(results, key=node_position))
 
 
 ###
 # Path expressions
 @method('//', bp=75)
-def nud_descendant_path(self):
+def nud_descendant_path(self: XPathToken) -> XPathToken:
     if self.parser.next_token.label not in self.parser.PATH_STEP_LABELS:
         self.parser.expected_next(*self.parser.PATH_STEP_SYMBOLS)
 
@@ -682,7 +612,7 @@ def nud_descendant_path(self):
 
 
 @method('/', bp=75)
-def nud_child_path(self):
+def nud_child_path(self: XPathToken) -> XPathToken:
     if self.parser.next_token.label not in self.parser.PATH_STEP_LABELS:
         try:
             self.parser.expected_next(*self.parser.PATH_STEP_SYMBOLS)
@@ -695,7 +625,7 @@ def nud_child_path(self):
 
 @method('//')
 @method('/')
-def led_child_or_descendant_path(self, left):
+def led_child_or_descendant_path(self: XPathToken, left: XPathToken) -> XPathToken:
     if left.symbol in ('/', '//', ':', '[', '$'):
         pass
     elif left.label not in self.parser.PATH_STEP_LABELS and \
@@ -710,7 +640,8 @@ def led_child_or_descendant_path(self, left):
 
 
 @method('/')
-def select_child_path(self, context=None):
+def select_child_path(self: XPathToken, context: ContextType = None) \
+        -> Iterator[ItemType]:
     """
     Child path expression. Selects child:: axis as default (when bind to '*' or '(name)').
     """
@@ -723,12 +654,12 @@ def select_child_path(self, context=None):
         if isinstance(context.document, DocumentNode):
             context.item = context.document
         elif context.root is None or isinstance(context.root.parent, ElementNode):
-            return  # A rooted subtree -> document root produce []
+            return  # No root or a rooted subtree -> document root produce []
         else:
-            context.item = context.root
+            context.item = context.root  # A fragment or a schema node
         yield from self[0].select(context)
     else:
-        items = set()
+        items: Set[ItemType] = set()
         for _ in context.inner_focus_select(self[0]):
             if not isinstance(context.item, XPathNode):
                 msg = f"Intermediate step contains an atomic value {context.item!r}"
@@ -740,23 +671,22 @@ def select_child_path(self, context=None):
                 elif result in items:
                     pass
                 elif isinstance(result, ElementNode):
-                    if result.elem not in items:
+                    if result.obj not in items:
                         items.add(result)
                         yield result
                 else:
                     items.add(result)
                     yield result
-                    if isinstance(context, XPathSchemaContext):
-                        self[1].add_xsd_type(result)
 
 
 @method('//')
-def select_descendant_path(self, context=None):
+def select_descendant_path(self: XPathToken, context: ContextType = None) \
+        -> Iterator[ItemType]:
     """Operator '//' is a short equivalent to /descendant-or-self::node()/"""
     if context is None:
         raise self.missing_context()
     elif len(self) == 2:
-        items = set()
+        items: Set[ItemType] = set()
         for _ in context.inner_focus_select(self[0]):
             if not isinstance(context.item, XPathNode):
                 raise self.error('XPTY0019')
@@ -768,22 +698,20 @@ def select_descendant_path(self, context=None):
                     elif result in items:
                         pass
                     elif isinstance(result, ElementNode):
-                        if result.elem not in items:
+                        if result.obj not in items:
                             items.add(result)
                             yield result
                     else:
                         items.add(result)
                         yield result
-                        if isinstance(context, XPathSchemaContext):
-                            self[1].add_xsd_type(result)
 
     else:
         if isinstance(context.document, DocumentNode):
             context.item = context.document
         elif context.root is None or isinstance(context.root.parent, ElementNode):
-            return  # A rooted subtree -> document root produce []
+            return  # No root or a rooted subtree -> document root produce []
         else:
-            context.item = context.root
+            context.item = context.root  # A fragment or a schema node
 
         items = set()
         for _ in context.iter_descendants():
@@ -793,12 +721,10 @@ def select_descendant_path(self, context=None):
                 elif result in items:
                     pass
                 elif isinstance(result, ElementNode):
-                    if result.elem not in items:
+                    if result.obj not in items:
                         items.add(result)
                 else:
                     items.add(result)
-                    if isinstance(context, XPathSchemaContext):
-                        self[0].add_xsd_type(result)
 
         yield from sorted(items, key=node_position)
 
@@ -806,26 +732,25 @@ def select_descendant_path(self, context=None):
 ###
 # Predicate filters
 @method('[', bp=80)
-def led_predicate(self, left):
+def led_predicate(self: XPathToken, left: XPathToken) -> XPathToken:
     self[:] = left, self.parser.expression()
     self.parser.advance(']')
     return self
 
 
 @method('[')
-def select_predicate(self, context=None):
+def select_predicate(self: XPathToken, context: ContextType = None) -> Iterator[ItemType]:
     if context is None:
         raise self.missing_context()
 
-    for _ in context.inner_focus_select(self[0]):
+    for _ in context.inner_focus_select(self[0], True):
         if (self[1].label in ('axis', 'kind test') or self[1].symbol == '..') \
                 and not isinstance(context.item, XPathNode):
             raise self.error('XPTY0020')
-        elif False and isinstance(context, XPathSchemaContext):
-            yield context.item
-            continue
 
-        predicate = [x for x in self[1].select(copy(context))]
+        predicate: Sequence[NumericType]
+        predicate = [x for x in cast(Iterator[NumericType], self[1].select(copy(context)))]
+
         if len(predicate) == 1 and isinstance(predicate[0], NumericProxy):
             if context.position == predicate[0]:
                 yield context.item
@@ -836,19 +761,19 @@ def select_predicate(self, context=None):
 ###
 # Parenthesized expressions
 @method('(', bp=100)
-def nud_parenthesized_expr(self):
+def nud_parenthesized_expr(self: XPathToken) -> XPathToken:
     self[:] = self.parser.expression(),
     self.parser.advance(')')
     return self
 
 
 @method('(')
-def evaluate_parenthesized_expr(self, context=None):
+def evaluate_parenthesized_expr(self: XPathToken, context: ContextType = None) -> Any:
     return self[0].evaluate(context)
 
 
 @method('(')
-def select_parenthesized_expr(self, context=None):
+def select_parenthesized_expr(self: XPathToken, context: ContextType = None) -> Iterator[Any]:
     return self[0].select(context)
 
 # XPath 1.0 definitions continue into module xpath1_functions
